--
-- large Y-shaped joiner
--

PREFABS.Joiner_y1 =
{
  file   = "joiner/y1.wad"

  prob   = 100
  theme  = "tech"

  where  = "seeds"
  shape  = "I"

  seed_w = 3
  seed_h = 2

  deep   = 16
  over   = 16

  delta_h  = 24
  nearby_h = 128
  can_flip = true
}

