--
-- Exit closet
--

PREFABS.Exit_closet3 =
{
  file  = "exit/closet3.wad"

  prob  = 20
  theme = "!tech"

  where  = "seeds"
  seed_w = 2
  seed_h = 2

  deep  =  16
  over  = -16

  x_fit = "frame"
}

