
-----------------------------------
--   FraggleScript Thing Names   --
-----------------------------------

-- This maps an entity number to a keyword used by the fragglescript
-- spawn() function.  For legacy, these keywords are defined in the
-- THINGS.H lump (which a script must include).

FRAGGLESCRIPT_THINGS =
{
  [5] = "BLUEKEYCARD"
  [6] = "YELLOWKEYCARD"
  [7] = "SPIDERMASTERMIND"
  [8] = "BACKPACK"
  [9] = "SHOTGUY"

  [10] = "SLOP"
  [12] = "SLOP2"
  [13] = "REDKEYCARD"
  [14] = "TELEPORTMAN"
  [15] = "DEADPLAYER"
  [16] = "CYBERDEMON"
  [17] = "ECELLPACK"
  [18] = "DEADTROOPER"
  [19] = "DEADSERGEANT"
  [20] = "DEADIMP"
  [21] = "DEADDEMON"
  [22] = "DEADCACO"
  [23] = "DEADLOSTSOUL"
  [24] = "BLOODPOOL1"
  [25] = "DEADCORPSE1"
  [26] = "TWITCHCORPSE3"
  [27] = "SKULLPOLE2"
  [28] = "SKULLPOLE1"
  [29] = "SKULLPILE"
  [30] = "TALLGRNPILLAR"
  [31] = "SHRTGRNPILLAR"
  [32] = "TALLREDPILLAR"
  [33] = "SHRTREDPILLAR"
  [34] = "CANDLE"
  [35] = "CANDELABRA"
  [36] = "HEARTCOLUMN"
  [37] = "SKULLCOLUMN"
  [38] = "REDSKULLKEY"
  [39] = "YELLOWSKULLKEY"
  [40] = "BLUESKULLKEY"
  [41] = "EVILEYE"
  [42] = "SKULLROCK"
  [43] = "GRAYTREE"
  [44] = "TALLBLUFIRESTICK"
  [45] = "TALLGRNFIRESTICK"
  [46] = "TALLREDFIRESTICK"
  [47] = "STALAGMITE"
  [48] = "TALLTECHPILLAR"
  [49] = "TWITCHCORPSE1"
  [50] = "HANGINGMAN1"
  [51] = "HANGINGMAN2"
  [52] = "HANGINGMAN3"
  [53] = "HANGINGMAN4"
  [54] = "TREE"
  [55] = "SHRTBLUFIRESTICK"
  [56] = "SHRTGRNFIRESTICK"
  [57] = "SHRTREDFIRESTICK"
  [58] = "SPECTRE"
  [59] = "HANGINGMAN5"
  [60] = "HANGINGMAN6"
  [61] = "HANGINGMAN7"
  [62] = "HANGINGMAN8"
  [63] = "TWITCHCORPSE2"
  [64] = "ARCHVILE"
  [65] = "CHAINGUY"
  [66] = "REVENANT"
  [67] = "MANCUBUS"
  [68] = "ARACHNOTRON"
  [69] = "HELLKNIGHT"
  [70] = "BURNINGBARREL"
  [71] = "PAINELEMENTAL"
  [72] = "KEEN"
  [73] = "HANGINGMAN9"
  [74] = "HANGINGMAN10"
  [75] = "HANGINGMAN11"
  [76] = "HANGINGMAN12"
  [77] = "HANGINGMAN13"
  [78] = "HANGINGMAN14"
  [79] = "BLOODPOOL2"
  [80] = "BLOODPOOL3"
  [81] = "BLOODPOOL4"
  [82] = "SUPERSHOTGUN"
  [83] = "MEGASPHERE"
  [84] = "WOLFSS"
  [85] = "TALLTECHLAMP"
  [86] = "SHORTTECHLAMP"
  [87] = "BOSSTARGET"
  [88] = "BOSSBRAIN"
  [89] = "BOSSSPIT"

  [888]  = "DOGS"
  [2001] = "SHOTGUN"
  [2002] = "CHAINGUN"
  [2003] = "RLAUNCHER"
  [2004] = "PLASMAGUN"
  [2005] = "CHAINSAW"
  [2006] = "BFG"
  [2007] = "CLIP"
  [2008] = "SHELLS"
  [2010] = "ROCKET"
  [2011] = "STIMPACK"
  [2012] = "MEDIKIT"
  [2013] = "SUPERCHARGE"
  [2014] = "HEALTHPOTION"
  [2015] = "ARMORHELMET"
  [2018] = "GREENARMOR"
  [2019] = "BLUEARMOR"
  [2022] = "INVULNERABILITY"
  [2023] = "BESERKPACK"
  [2024] = "INVISIBILITY"
  [2025] = "RADSUIT"
  [2026] = "AUTOMAP"
  [2028] = "FLOORLAMP"
  [2035] = "BARREL"
  [2045] = "LITEAMP"
  [2046] = "ROCKETBOX"
  [2047] = "ECELL"
  [2048] = "BULLETBOX"
  [2049] = "SHELLBOX"
  [3001] = "IMP"
  [3002] = "DEMON"
  [3003] = "BARONOFHELL"
  [3004] = "TROOPER"
  [3005] = "CACODEMON"
  [3006] = "LOSTSOUL"

  [5001] = "PUSH"
  [5002] = "PULL"
  [5003] = "CAMERA"
  [5004] = "NODE"
}

