--
-- Simple start closet (with door)
--

PREFABS.Start_closet1 =
{
  file  = "start/closet1.wad"
  map   = "MAP01"

  prob  = 80

  where = "seeds"
  deep  =  16
  over  = -16

  x_fit = "frame"
  y_fit = "top"
}

