--
-- Closet with a drop-off
--

PREFABS.Start_Dropoff =
{
  file   = "start/dropoff.wad"

  prob   = 80

  where  = "seeds"

  seed_w = 2
  deep   =  16
  over   = -16

  height = 128

  x_fit  = "frame"
  y_fit  = "top"

  nearby_h = 160
}

