--
-- Water Fountain
--


PREFABS.Decor_water_fountain =
{
  file   = "decor/fountain.wad"
  map    = "MAP01"

  prob   = 1000
  theme  = "urban"
  env    = "outdoor"

  where  = "point"
  size   = 128

  bound_z1 = 0

  sink_mode = "never"
}

