--
-- Hanging chandelier
--

PREFABS.Decor_chandelier =
{
  file   = "decor/chandelier.wad"
  map    = "MAP01"

  prob   = 100
  env    = "building"

  kind   = "light"
  where  = "point"

  height = 160

  bound_z1 = -64
  bound_z2 = 0
}

