--
-- basic stair, 64 units high
--

PREFABS.Stair_64_red =
{
  file   = "stairs/stair_64.wad"

  prob  = 10
  theme = "urban"

  where  = "seeds"
  shape  = "I"

  seed_w = 2

  x_fit  = "stretch"

  bound_z1 = 0

  delta_h = 64
}

