--
-- Carpets (flat floor designs)
--

PREFABS.Decor_techfloor1 =
{
  file   = "decor/carpets.wad"
  map    = "MAP01"

  prob   = 50
  theme  = "tech"
  env    = "building"

  where  = "point"
  size   = 80

  bound_z1 = 0
  bound_z2 = 32

  sink_mode = "never"
}


PREFABS.Decor_sunkgrate =
{
  file   = "decor/carpets.wad"
  map    = "MAP02"

  prob   = 50
  theme  = "urban"
  env    = "building"

  where  = "point"
  size   = 80

  bound_z1 = 0
  bound_z2 = 32

  sink_mode = "never"
}

