--
-- Joiner with opening door
--

PREFABS.Joiner_door2 =
{
  file   = "joiner/door2.wad"
  map    = "MAP01"

  prob   = 100
  style  = "doors"

  where  = "seeds"
  shape  = "I"

  seed_w = 2
  seed_h = 1

  deep   = 16
  over   = 16

  x_fit  = "frame"
}

