--
-- Hellish pentagram
--

PREFABS.Item_pentagram =
{
  file  = "item/pentagram.wad"

  rank  = 2
  prob  = 100
  theme = "hell"

  item_kind = "key"

  where = "point"
  size  = 90   -- actually bigger, but won't block movement
}

