--
--  Small closet with a door
--

PREFABS.Item_closet_w_door =
{
  file   = "item/closet.wad"

  prob   = 15
  env    = "!cave"

  where  = "seeds"
  seed_w = 1
  seed_h = 1

  deep =  16
  over = -16

  x_fit = "frame"
  y_fit = "top"
}

