--
-- very cool Y-shapish 3x2 joiner 
--

PREFABS.Joiner_y3 =
{
  file   = "joiner/y3.wad"
  map    = "MAP01"

  theme  = "hell"
  prob   = 150

  where  = "seeds"
  shape  = "I"

  seed_w = 3
  seed_h = 2

  deep   = 16
  over   = 16

  delta_h  = 32
  nearby_h = 160
  can_flip = true
}

