------------------------------------------------------------------------
--  HERETIC PARAMETERS and ACTIONS
------------------------------------------------------------------------
--
--  Oblige Level Maker
--
--  Copyright (C) 2006-2017 Andrew Apted
--
--  This program is free software; you can redistribute it and/or
--  modify it under the terms of the GNU General Public License
--  as published by the Free Software Foundation; either version 2
--  of the License, or (at your option) any later version.
--
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--  GNU General Public License for more details.
--
------------------------------------------------------------------------

HERETIC.PARAMETERS =
{
  teleporters = true

  jump_height = 24

  max_name_length = 28

  titlepic_lump   = "TITLE"
  titlepic_format = "raw"

  skip_monsters = { 10,20 }

  monster_factor = 1.0
  health_factor  = 1.0
  ammo_factor    = 1.0
  time_factor    = 1.0

  mon_along_factor = 7.2
}


HERETIC.ACTIONS =
{
  --
  -- These keywords are used by prefabs that are remotely
  -- triggered (by a switch or walk-over line).
  --

  S1_OpenDoor = { id=103,  kind="open" }    -- opens and stays open
  W1_OpenDoor = { id=2,    kind="open" }    --
  GR_OpenDoor = { id=46,   kind="open" }    --

  W1_OpenDoorFast = { id=2, kind="open" }   -- [ Heretic lacks this ]

  S1_RaiseStair = { id=106,  kind="stair" }  -- 16 units
  W1_RaiseStair = { id=107,  kind="stair" }  --

  S1_FloorUp    = { id=18,  kind="floor_up" } -- up to next highest floor

  S1_LowerFloor = { id=23, kind="lower" }  -- down to lowest nb floor
  W1_LowerFloor = { id=38, kind="lower" }  --
}

