--
-- Liquid arch
--

UNFINISHED.Window_liquid_arch =
{
  file   = "window/liq_arch.wad"

  prob   = 90

  where  = "edge"
  deep   = 16
  over   = 16

  x_fit  = "frame"

  bound_z1 = 0
  bound_z2 = 128
}

