--
-- curved stair, 32 units high
--

PREFABS.Stair_curve_32 =
{
  file   = "stairs/curve_32.wad"

  prob   = 50

  where  = "seeds"
  shape  = "L"

  bound_z1 = 0
  bound_z2 = 32

  delta_h = 32
  plain_ceiling = true
}

