--
-- Joiner with remote door
--

PREFABS.Joiner_remote_door =
{
  file   = "joiner/barred2.wad"
  where  = "seeds"
  shape  = "I"

  key    = "barred"

  prob   = 20

  seed_w = 2
  seed_h = 1

  deep   = 16
  over   = 16

  tag_1  = "?door_tag"
  door_action = "S1_OpenDoor"

  flat_FLAT23 = "SILVER3"
}


PREFABS.Joiner_remote_sw_metal =
{
  template = "Joiner_remote_door"

  key = "sw_metal"

  x_fit = "frame"
  y_fit = "stretch"

  tex_SILVER3 = "METAL3"
  flat_FLAT23 = "CEIL5_2"
}

