--
-- huge joiner with a round glowy thing
--

-- disabled for now, since connected area is almost always too low

UNFINISHED.Joiner_y2 =
{
  file   = "joiner/y2.wad"

  prob   = 30
  theme  = "!hell"
  env    = "!cave"

  where  = "seeds"
  shape  = "I"

  seed_w = 3
  seed_h = 2

  deep   = 16
  over   = 16

  delta_h  = 32
  nearby_h = 200
  can_flip = false
}

