--
-- Small Cage
--

PREFABS.Cage_small_1 =
{
  file   = "cage/small_cage.wad"
  map    = "MAP01"

  prob   = 50
  env    = "outdoor"

  where  = "point"

  size   = 80
  height = 128

  bound_z1 = 0
  bound_z2 = 120
}

