--
-- A huge pillar
--

PREFABS.Pillar_tech3 =
{
  file   = "decor/pillar3.wad"
  map    = "MAP01"

  rank   = 2
  prob   = 100
  theme  = "tech"
  env    = "building"

  where  = "point"
  size   = 200
  height = 192

  bound_z1 = 0
  bound_z2 = 192

  z_fit  = { 80,112 }

  sink_mode = "never"
}

