
test_wad = {
    "theme": "bit_mixed",
    "size": "micro",
    "outdoors": "none",
    "caves": "none",
    "liquids": "none",
    "hallways": "none",
    "teleporters": "none",
    "steepness": "none",

    "mons": "none",
    "strength": "weak",
    "ramp_up": "slow",
    "bosses": "none",
    "traps": "none",
    "cages": "none",

    "health": "none",
    "ammo": "none",
    "weapons": "none",
    "items": "none",
    "secrets": "none",

    "misc": 1,
    # misc module options:
    "pistol_starts": "yes",
    "alt_starts": "no",
    "big_rooms": "none",
    "parks": "none",
    "windows": "none",
    "symmetry": "none",
    "darkness": "none",
    "mon_variety": "none",
    "barrels": "none",
    "doors": "none",
    "keys": "none",
    "switches": "none"

}

childs_play_wad = {
    "theme": "bit_mixed",
    "size": "micro",
    "outdoors": "none",
    "caves": "mixed",
    "liquids": "mixed",
    "hallways": "none",
    "teleporters": "none",
    "steepness": "none",

    "mons": "scarce",
    "strength": "weak",
    "ramp_up": "slow",
    "bosses": "none",
    "traps": "none",
    "cages": "none",

    "health": "normal",
    "ammo": "normal",
    "weapons": "normal",
    "items": "none",
    "secrets": "none",

    "misc": 1,
    # misc module options:
    "pistol_starts": "yes",
    "alt_starts": "no",
    "big_rooms": "none",
    "parks": "none",
    "windows": "rare",
    "symmetry": "rare",
    "darkness": "rare",
    "mon_variety": "rare",
    "barrels": "rare",
    "doors": "none",
    "keys": "none",
    "switches": "none"
}
