//------------------------------------------------------------------------
//  UI INCLUDES
//------------------------------------------------------------------------
//
//  Oblige Level Maker
//
//  Copyright (C) 2006-2009 Andrew Apted
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//------------------------------------------------------------------------

#ifndef __OB_HDR_UI_H__
#define __OB_HDR_UI_H__

#include "zf_menu.h"

#include "ui_hyper.h"
#include "ui_map.h"
#include "ui_rchoice.h"

#include "ui_build.h"
#include "ui_game.h"
#include "ui_module.h"
#include "ui_window.h"

#endif /* __OB_HDR_UI_H__ */

