--
-- Hell-ish medium size cage
--

UNFINISHED.Cage_hellbox1 =
{
  file   = "cage/hell_box.wad"
  map    = "MAP01"

  rank   = 2
  prob   = 100
  theme  = "hell"
  env    = "building"

  where  = "point"
  size   = 128
  height = { 160,256 }

  bound_z1 = 0
  bound_z2 = 160

  z_fit = "frame"
}

