--
-- Waterfall for caves
--

UNFINISHED.Decor_cave_waterfall =
{
  file   = "decor/waterfall.wad"

  prob   = 100
  env    = "cave"

  where  = "seeds"
  shape  = "U"

  seed_w = 3
  seed_h = 1

  deep   = 192

  y_fit  = "top"
}

