--
--  Niche switch, blocked by raising bars
--

PREFABS.Switch_niche_bars =
{
  file   = "switch/niche_bars.wad"

  prob   = 50
  key    = "barred"

  where  = "seeds"

  seed_w = 1
  seed_h = 1

  tag_1 = "?switch_tag"
  tag_2 = "?door_tag"
}

