--
-- Arch joiners
--

PREFABS.Joiner_archy1 =
{
  file   = "joiner/archy.wad"
  map    = "MAP01"

  prob  = 140

  where  = "seeds"
  shape  = "I"

  seed_w = 2
  seed_h = 1

  deep   = 16
  over   = 16

  y_fit = "stretch"
}


PREFABS.Joiner_archy1_wide =
{
  file   = "joiner/archy.wad"
  map    = "MAP02"

  prob  = 400

  where  = "seeds"
  shape  = "I"

  seed_w = 3
  seed_h = 1

  deep   = 16
  over   = 16

  x_fit = "frame"
}

