--
--  Niche switch
--

PREFABS.Switch_niche2 =
{
  file   = "switch/niche2.wad"
  map    = "MAP01"

  key    = "sw_metal"
  prob   = 50

  where  = "seeds"
  deep   =  16
  over   = -16

  x_fit = "frame"
  y_fit = "top"

  tag_1 = "?switch_tag"
}

