--
-- Cage which is flush with the wall
--

PREFABS.Cage_wall =
{
  file   = "cage/wall_cage.wad"
  map    = "MAP01"

  prob  = 100

  where  = "seeds"
  shape  = "U"

  seed_w = 1
  seed_h = 1

  deep   =  16
  over   = -16

  x_fit = "stretch"
  y_fit = "top"
}

