--
-- Ruins (outdoor shapes)
--

PREFABS.Decor_ruins1 =
{
  file   = "decor/ruins.wad"
  map    = "MAP01"

  prob   = 50
  theme  = "urban"
  env    = "outdoor"

  where  = "point"
  size   = 80

  bound_z1 = 0
  bound_z2 = 128
}


PREFABS.Decor_ruins2 =
{
  template = "Decor_ruins1"
  map      = "MAP02"
}


PREFABS.Decor_ruins3 =
{
  template = "Decor_ruins1"
  map      = "MAP03"
}

