--
-- Hanging ceiling lamp
--

PREFABS.Light_hanging =
{
  file   = "decor/hang_lamp.wad"
  map    = "MAP01"

  prob   = 100
  theme  = "urban"
  env    = "building"

  kind   = "light"
  where  = "point"

  height = 160

  bound_z1 = -64
  bound_z2 = 0
}

