--
-- Joiner with stairs
--

PREFABS.Joiner_stair1 =
{
  file   = "joiner/stair1.wad"

  prob   = 60
  theme  = "urban"
  style  = "steepness"

  where  = "seeds"
  shape  = "I"

  seed_w = 2
  seed_h = 1

  deep   = 16
  over   = 16

  x_fit  = "stretch"

  delta_h  = 48
  nearby_h = 160
  can_flip = true
}

