--
--  Niche switch
--

PREFABS.Switch_wall_box1 =
{
  file   = "switch/wall_box1.wad"

  prob   = 50

  where  = "seeds"
  seed_w = 1
  seed_h = 1

  deep   =  16
  over   = -16

  x_fit = "frame"
  y_fit = "top"

  tag_1 = "?switch_tag"
}

