--
-- Simple joiner II
--

PREFABS.Joiner_simple2 =
{
  file   = "joiner/simple2.wad"
  map    = "MAP01"

  theme  = "!tech"
  prob   = 1400

  where  = "seeds"
  shape  = "I"

  seed_w = 3
  seed_h = 1

  deep   = 16
  over   = 16

  x_fit = "frame"
--y_fit = "frame"
}

