--
-- Derek's indoor techy computer thing
--

PREFABS.Decor_dittohead_computer =
{
  file   = "decor/dh_comp.wad"
  map    = "MAP01"

  prob   = 5000
  theme  = "tech"
  env    = "building"

  where  = "point"
  size   = 128
  height = 160

  bound_z1 = 0
  bound_z2 = 160

  z_fit  = { 128,136 }
}

