--
--  Small item alcove
--

PREFABS.Item_alcove_small =
{
  file  = "item/alcove1.wad"

  prob  = 50

  where  = "seeds"
  seed_w = 1
  seed_h = 1

  deep =  16
  over = -16

  x_fit = "frame"
  y_fit = "top"
}

