--
--  Large gothic item balcony
--

PREFABS.Item_gothic_balcony =
{
  file   = "item/gothic1.wad"

  rank   = 3
  prob   = 200
  theme  = "!tech"
  env    = "!cave"

  item_kind = "key"

  where  = "seeds"
  seed_w = 3
  seed_h = 2

  deep =  16
  over = -16

  x_fit = "frame"
}

