--
-- vent piece : corner (L shape)
--

PREFABS.Hallway_vent_c1 =
{
  file   = "hall/vent_c.wad"
  map    = "MAP01"

  group  = "vent"
  prob   = 50

  where  = "seeds"
  shape  = "L"

  mon_height = 64
}

