--
-- long narrow stair, 2 seeds high
--

PREFABS.Stair_long_1x2 =
{
  file   = "stairs/long_1x2.wad"

  prob   = 90

  where  = "seeds"
  shape  = "I"
  
  seed_h = 2

  x_fit  = "stretch"

  bound_z1 = 0
  bound_z2 = 64

  delta_h = 64
}

