--
--  Medium techy alcove for keys
--

PREFABS.Item_alcove_techy =
{
  file   = "item/alcove2.wad"

  rank   = 2
  prob   = 200
  theme  = "tech"
  env    = "!cave"

  item_kind = "key"

  where  = "seeds"
  seed_w = 2
  seed_h = 2

  deep =  16
  over = -16

  x_fit = "frame"
  y_fit = "top"
}

