--
-- Double lighting niche
--

PREFABS.Wall_lite2_blue =
{
  file   = "wall/lite_2.wad"
  map    = "MAP01"

  prob   = 50
  group  = "lite2"

  where  = "edge"
  height = 128
  deep   = 16

  x_fit  = "frame"
  z_fit  = "frame"

  bound_z1 = 0
  bound_z2 = 128
}

