--
-- 2-seed-wide hallway : corner (L shape)
--

PREFABS.Hallway_deuce_c1 =
{
  file   = "hall/deuce_c.wad"
  map    = "MAP01"

  group  = "deuce"
  prob   = 50

  where  = "seeds"
  shape  = "L"

  seed_w = 2
  seed_h = 2
}

