--
-- Another outdoor lamp
--

PREFABS.Decor_blob_lamp =
{
  file   = "decor/blob_lamp.wad"
  map    = "MAP01"

  prob   = 150
  skip_prob = 75

  theme  = "!hell"
  env    = "outdoor"
  game   = "doom2"

  where  = "point"
  size   = 80
}

