--
-- Standing Cage
--

PREFABS.Cage_stand1 =
{
  file   = "cage/stand_cage.wad"
  map    = "MAP01"

  prob   = 500

  where  = "point"

  size   = 128
  height = 128

  bound_z1 = 0
  bound_z2 = 120
}


PREFABS.Cage_stand2 =
{
  file   = "cage/stand_cage.wad"
  map    = "MAP02"

  rank   = 2
  prob   = 100
  env    = "cave"

  where  = "point"

  size   = 96
  height = 152

  bound_z1 = 0
  bound_z2 = 152
}

