--
-- Teleporter pad
--

PREFABS.Teleporter1 =
{
  file   = "teleporter/pad1.wad"
  map    = "MAP01"

  theme  = "!tech"
  prob   = 50

  where  = "point"

  tag_1 = "?out_tag"
  tag_2 = "?in_tag"
}

