--
--  Secret item masquerading as a cage
--

PREFABS.Item_secret_cage =
{
  file  = "item/secret_cage.wad"
  where = "seeds"

  prob  = 200
  env   = "building"

  key   = "secret"

  seed_w = 2
  seed_h = 2
  height = 120

  deep =  16
  over = -16

  x_fit = "frame"
}

