--
-- Trap looking like a normal switch
--

PREFABS.Cage_fake_switch =
{
  file   = "cage/fake_switch.wad"

  prob   = 20
  style  = "traps"

  where  = "seeds"
  shape  = "U"

  seed_w = 2
  seed_h = 1

  deep   =  16
  over   = -16

  y_fit  = "top"

  bound_z1 = 0
}

