--
--  Outdoor item gazebo
--

PREFABS.Item_gazebo2 =
{
  file   = "item/gazebo2.wad"

  rank   = 2
  prob   = 100

  env    = "outdoor"
  open_to_sky = 1
  height = 176

  item_kind = "key"

  where  = "seeds"
  seed_w = 2
  seed_h = 2

  x_fit = "frame"
}

