--
-- vent piece : four-way "+" junction
--

PREFABS.Hallway_vent_p1 =
{
  file   = "hall/vent_p.wad"
  map    = "MAP01"

  group  = "vent"
  prob   = 50

  where  = "seeds"
  shape  = "P"

  mon_height = 64
}

