--
--  Exit switch for Heretic
--

PREFABS.Exit_heretic =
{
  file   = "exit/sw_heretic.wad"
  map    = "MAP01"

  prob   = 5

  where  = "point"
}


PREFABS.Exit_heretic_secret =
{
  file   = "exit/sw_heretic.wad"
  map    = "MAP02"

  prob   = 5

  kind   = "secret_exit"
  where  = "point"
}

