--
-- Teleporter closet
--

PREFABS.Teleporter_closet =
{
  file   = "teleporter/closet2.wad"

  prob   = 50

  where  = "seeds"
  seed_w = 1
  seed_h = 1

  deep  =  16
  over  = -16

  x_fit = "frame"
  y_fit = "top"

  tag_1 = "?out_tag"
  tag_2 = "?in_tag"
}

