--
--  Secret item niche
--

-- the hint here is misaligned texture
PREFABS.Item_secret_niche1 =
{
  file  = "item/secret_niche.wad"
  map   = "MAP01"

  prob  = 100
  key   = "secret"
  where = "seeds"

  seed_w = 1
  seed_h = 1

  deep =  16
  over = -16

  x_fit = "frame"
  y_fit = "top"
}

