--
-- Secret plain fence
--

PREFABS.Fence_plain_secret =
{
  file   = "fence/secret_16.wad"
  map    = "MAP06"

  prob   = 50
  key    = "secret"

  where  = "edge"
  seed_w = 2

  deep   = 16
  over   = 16

  x_fit  = "frame"

  fence_h  = 32
  bound_z1 = 0
}

