--
-- L-shape Joiner with stairs
--

PREFABS.Joiner_curve1 =
{
  file   = "joiner/curve1.wad"

  prob   = 100
  style  = "steepness"

  where  = "seeds"
  shape  = "L"

  seed_w = 2
  seed_h = 2

  delta_h  = 64
  nearby_h = 128
}

