--
-- vent piece : dead-end
--

PREFABS.Hallway_vent_u1 =
{
  file   = "hall/vent_u.wad"
  map    = "MAP01"

  group  = "vent"
  prob   = 50

  where  = "seeds"
  shape  = "U"
}


PREFABS.Hallway_vent_cage1 =
{
  file   = "hall/vent_u.wad"
  map    = "MAP02"

  group  = "vent"
  style  = "cages"
  prob   = 50

  where  = "seeds"
  shape  = "U"
}

