--
-- Joiner with lift
--

PREFABS.Joiner_lift3 =
{
  file   = "joiner/lift3.wad"

  prob   = 5
  style  = "steepness"

  env      = "!cave"
  neighbor = "!cave"

  where  = "seeds"
  shape  = "I"

  seed_w = 2
  seed_h = 1

  deep   = 16
  over   = 16

  x_fit  = "frame"

  delta_h  = 104
  can_flip = true
}

