--
-- An open start closet
--

PREFABS.Start_closet2 =
{
  file  = "start/closet2.wad"
  map   = "MAP01"

  prob  = 240

  where  = "seeds"
  seed_w = 2

  deep  =  16
  over  = -16

  x_fit = "frame"
  y_fit = "top"
}

